import numpy as np
import itertools
import math
import multiprocess
from collections import defaultdict


def bravo(x, mu_alt, num_samples=None):
    num_ones = np.count_nonzero(x == 1)
    num_zeros = np.count_nonzero(x == 0)
    num_halves = np.count_nonzero(x == 1 / 2)

    num_samples = len(x) if num_samples is None else num_samples

    x_contribution = np.ones(num_samples)

    n = len(x)

    for i in range(num_samples):
        random_sample = np.random.choice(
            (1, 0, 1 / 2), p=[num_ones / n, num_zeros / n, num_halves / n]
        )

        if random_sample == 1:
            x_contribution[i] = 2 * mu_alt
        elif random_sample == 0:
            x_contribution[i] = 2 * (1 - mu_alt)
        elif random_sample == 1 / 2:
            # do nothing to martingale, but remove that observation from
            # the population.
            num_halves -= 1
            n -= 1

    martingale = np.cumprod(x_contribution)

    return martingale


def get_bravo_workload(x, mu_alt, alpha, num_samples=None):
    num_samples = len(x) if num_samples is None else num_samples

    unique_ballots = set()
    ballot_numbers = list(range(len(x)))
    sample_num = 1
    martingale = 1
    while martingale < 1 / alpha and len(unique_ballots) < num_samples:
        random_ballot_number = np.random.choice(ballot_numbers)

        unique_ballots.add(random_ballot_number)

        ballot = x[random_ballot_number]
        martingale *= (
            2 * mu_alt if ballot == 1 else 2 * (1 - mu_alt) if ballot == 0 else 1
        )
        sample_num += 1

    workload = len(unique_ballots)

    return workload


def get_workload_from_mart(x, mart_fn, alpha):
    where_cross = np.where(mart_fn(x) >= 1 / alpha)[0]
    if len(where_cross) == 0:
        workload = len(x)
    else:
        workload = where_cross[0]

    return workload


def get_data_dict(N, margins, nuisances):
    election_outcomes_dicts = []

    for margin, nuisance in itertools.product(margins, nuisances):
        election_outcomes_dicts = election_outcomes_dicts + [
            {"margin": margin, "nuisances": nuisance}
        ]

    data_dict = {}
    for election_outcomes_dict in election_outcomes_dicts:
        halves = election_outcomes_dict["nuisances"]
        ones = int((N - halves) * (0.5 + election_outcomes_dict["margin"]))
        zeros = int(N - halves - ones)
        data = np.hstack(
            (
                np.repeat(0, zeros),
                np.repeat(1, ones),
                np.repeat(0.5, halves),
            )
        )

        name = (
            r"$(N_A^\star, N_U^\star, N_B^\star) = ("
            + str(ones)
            + ", "
            + str(halves)
            + ", "
            + str(zeros)
            + ")$"
        )
        data_dict[name] = np.random.permutation(data)

    return data_dict


def get_workloads(workload_dict, data, nsim=100, alpha=0.05, num_proc=1):
    """
    Get workloads for a collection of audits

    Parameters
    ----------
    workload_dict, dict of {string : function}
        dictionary of various functions where the key is a
        string indicating the name of the audit, and
        the function is a univariate function which takes
        an array-like of [0, 1]-bounded real numbers (the
        observations) and outputs the workload for the audit.

    data, array-like of [0, 1]-bounded reals
        The population of ballots

    nsim, positive integer
        number of simulations to perform

    alpha, (0, 1)-bounded real
        Significance level

    num_proc, positive integer
        Number of CPU processes to spawn. This should be less than or
        equal to the number of CPU cores on your machine.
    """
    N = np.sum(len(data))

    # Get things ready for the plot
    t = np.ones(N).cumsum()

    audit_names = list(workload_dict.keys())
    stopping_times_dict = {}

    def get_workload(i):
        np.random.seed()
        np.random.shuffle(data)

        result_array = [None] * len(audit_names)
        for i in range(len(audit_names)):
            audit_name = audit_names[i]
            workload_closure = workload_dict[audit_name]
            workload_value = workload_closure(data)
            workload_value = np.minimum(workload_value, N)

            result_array[i] = workload_value
        return result_array

    with multiprocess.Pool(processes=num_proc) as pool:
        results = np.array(pool.map(get_workload, range(nsim)))

    for j in range(len(audit_names)):
        audit_name = audit_names[j]
        stopping_times_dict[audit_name] = results[:, j]

    return stopping_times_dict



from typing import Callable
import numpy as np
from confseq.betting import (
    betting_mart,
    diversified_betting_mart,
    mu_t,
)
from rilacs.strategies import (
    apriori_Kelly_bet,
    square_gamma_dist,
    uniform_gamma_dist,
    get_conv_weights_from_dist,
)


def apriori_Kelly_martingale(
    x: np.ndarray,
    N: int,
    n_A: int,
    n_B: int,
    m: float = 1 / 2,
) -> np.ndarray:
    """The "a priori Kelly" martingale

    This function implements the "a priori Kelly" martingale from RiLACS
    (https://arxiv.org/pdf/2107.11323.pdf). This martingale attempts to
    construct an effective lambda value (a tuning parameter) based on the
    reported outcomes n_A and n_B for candidates A and B, respectively.

    Parameters
    ----------
    x : np.ndarray
        A numpy array containing a sequence of ballots with 1 denoting votes for
        candidate A, 0 for candidate B, and 1/2 for nuisance/invalid ballots.
    N : int
        The total number of ballots cast.
    n_A : int
        The reported number of ballots for candidate A
    n_B : int
        The reported number of ballots for candidate B.
    m : float
        The "null hypothesis" mean of x. For example, to test the null that
        candidate A ballots <= candidate B ballots, set this to 1/2.

    Returns
    -------
    np.ndarray
        The resulting nonnegative martingale
    """
    apriori_Kelly_lambda = lambda x, m: apriori_Kelly_bet(n_A, n_B)
    return betting_mart(
        x,
        m,
        lambdas_fn_positive=apriori_Kelly_lambda,
        N=N,
        convex_comb=True,
        theta=1,
        trunc_scale=1,
        m_trunc=True,
    )


def distKelly_martingale(
    x: np.ndarray,
    N: int,
    dist: Callable[[float], float],
    m: float = 1 / 2,
    D: int = 10,
    beta: float = 1,
) -> np.ndarray:
    """The distKelly martingale

    This function implements the distKelly martingale from RiLACS
    (https://arxiv.org/pdf/2107.11323.pdf). This method uses `D`
    lambda values, constructs `D` individual martingales from them,
    and then takes a convex combination of them according to the distribution
    specified in `dist`.

    Parameters
    ----------
    x : np.ndarray
        A numpy array containing a sequence of ballots.
    N : int
        The total number of ballots cast.
    dist : Callable[[float], float]
        The function from [0, 1] -> R describing the weights theta_1, ..., theta_D
    m : float
        The "null hypothesis" mean of x. In other words, distKelly forms a
        nonnegative martingale starting at one when mean(x) = m.
    D : int
        The size of the grid to derive lambda values from.
    beta : float
        Convex weights describing the alternatives to have power against.
        For example, beta = 1 corresponds to a one-sided test of mean(x) <= m,
        while beta = 0 corresponds to a one-sided test of mean(x) >= m, and
        beta = 1/2 yields a two-sided test.


    Returns
    -------
    np.ndarray
        The resulting nonnegative martingale
    """
    lambdas_fns_positive = [
        lambda x, m, i=i: (i + 1) / (mu_t(x, m, N) * (D + 1)) for i in range(D)
    ]
    lambdas_fns_negative = [
        lambda x, m, i=i: (i + 1) / ((1 - mu_t(x, m, N)) * (D + 1)) for i in range(D)
    ]

    lambdas_weights = get_conv_weights_from_dist(dist=dist, D=D)

    with np.errstate(divide="ignore"):
        mart = diversified_betting_mart(
            x,
            m,
            lambdas_fns_positive=lambdas_fns_positive,
            lambdas_fns_negative=lambdas_fns_negative,
            lambdas_weights=lambdas_weights,
            N=N,
            theta=beta,
            trunc_scale=1,
            m_trunc=True,
            convex_comb=True,
        )
    return mart


def sqKelly_martingale(
    x: np.ndarray,
    m: float,
    N: int,
    D: int = 10,
    beta: float = 1,
) -> np.ndarray:
    """The SqKelly martingale

    Instantiation of distKelly_martingale with dist=square_gamma_dist.
    """
    return distKelly_martingale(x=x, m=m, N=N, dist=square_gamma_dist, D=D, beta=beta)


def dKelly_martingale(
    x: np.ndarray,
    m: float,
    N: int,
    D: int = 10,
    beta: float = 1,
) -> np.ndarray:
    """The dKelly martingale

    Instantiation of distKelly_martingale with dist=uniform_gamma_dist.
    """
    return distKelly_martingale(x=x, m=m, N=N, dist=uniform_gamma_dist, D=D, beta=beta)



from typing import Callable, Tuple
import numpy as np
from rilacs.strategies import square_gamma_dist, uniform_gamma_dist
from rilacs.martingales import apriori_Kelly_martingale, distKelly_martingale
from confseq.betting import cs_from_martingale


def apriori_Kelly(
    x: np.ndarray,
    N: int,
    n_A: int,
    n_B: int,
    breaks: int = 1000,
    alpha: float = 0.05,
    running_intersection: bool = True,
    parallel: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    return cs_from_martingale(
        x,
        mart_fn=lambda x, m: apriori_Kelly_martingale(x, m=m, N=N, n_A=n_A, n_B=n_B),
        breaks=breaks,
        alpha=alpha,
        N=N,
        running_intersection=running_intersection,
        parallel=parallel,
    )


def distKelly(
    x: np.ndarray,
    N: int,
    dist: Callable[[float], float],
    D: int = 10,
    beta: float = 1,
    breaks: int = 1000,
    alpha: float = 0.05,
    running_intersection: bool = True,
    parallel: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    return cs_from_martingale(
        x,
        mart_fn=lambda x, m: distKelly_martingale(
            x, m=m, N=N, dist=dist, D=D, beta=beta
        ),
        breaks=breaks,
        alpha=alpha,
        N=N,
        running_intersection=running_intersection,
        parallel=parallel,
    )


def sqKelly(
    x: np.ndarray,
    N: int,
    D: int = 10,
    beta: float = 1,
    breaks: int = 1000,
    alpha: float = 0.05,
    running_intersection: bool = True,
    parallel: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    return distKelly(
        x,
        N=N,
        dist=square_gamma_dist,
        D=D,
        beta=beta,
        breaks=breaks,
        alpha=alpha,
        running_intersection=running_intersection,
        parallel=parallel,
    )


def dKelly(
    x: np.ndarray,
    N: int,
    D: int = 10,
    beta: float = 1,
    breaks: int = 1000,
    alpha: float = 0.05,
    running_intersection: bool = True,
    parallel: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    return distKelly(
        x,
        N=N,
        dist=uniform_gamma_dist,
        D=D,
        beta=beta,
        breaks=breaks,
        alpha=alpha,
        running_intersection=running_intersection,
        parallel=parallel,
    )



from typing import Callable
import numpy as np
from scipy.optimize import root


def apriori_Kelly_bet_general(x_reported: np.ndarray) -> float:
    objective = lambda l: np.sum((x_reported - 0.5) / (1 + l * (x_reported - 0.5)))

    sol = root(objective, x0=0.1)
    assert sol["success"]
    bet = sol["x"]
    return float(bet)


def apriori_Kelly_bet(n_A: int, n_B: int) -> float:
    return 2 * (n_A - n_B) / (n_A + n_B)


def linear_gamma_dist(y: float) -> float:
    return np.maximum(1 - 4 * y, 0)


def square_gamma_dist(y: float) -> float:
    return (y <= 1 / 3) * (1 / 3 - y) ** 2


def uniform_gamma_dist(y: float) -> float:
    return 1


def get_conv_weights_from_dist(dist: Callable[[float], float], D):
    unnormalized_weights = [dist((i + 1) / D) for i in range(D)]
    weights = np.array(unnormalized_weights) / sum(unnormalized_weights)
    return weights



import matplotlib.pyplot as plt
import time
import numpy as np
from logging import info
from rilacs.misc import get_workloads


def plot_workload(
    workload_dict,
    data_dict,
    nsim=100,
    alpha=0.05,
    num_proc=1,
    multiple_of=1,
    figsize=(3.5, 2.75),
    bbox_to_anchor=None,
    ncol=None,
    filename_prefix="stopping_times_",
):
    for data_name in data_dict:
        start_time = time.time()

        plt.style.use("seaborn-white")
        plt.style.use("seaborn-colorblind")
        plt.rcParams["font.family"] = "serif"
        fig, ax = plt.subplots(1, 1, figsize=figsize)

        for spine in ax.spines.values():
            spine.set_edgecolor("lightgrey")

        data = data_dict[data_name]
        stopping_times_dict = get_workloads(
            workload_dict=workload_dict,
            data=data,
            nsim=nsim,
            alpha=alpha,
            num_proc=num_proc,
        )

        # get counts of each of 0, 0.5, and 1. To use barplot in matplotlib
        u, inv = np.unique(data, return_inverse=True)
        counts = np.bincount(inv)
        for mart_name in workload_dict:
            stopping_times = stopping_times_dict[mart_name]
            avg_stopping_time_str = str(
                np.around(np.median(stopping_times), decimals=1)
            )

            ax.hist(
                stopping_times,
                alpha=0.7,
                label=mart_name + "\n(Median=" + avg_stopping_time_str + ")",
                histtype="step",
                density=True,
                linewidth=1.5,
            )

        ax.set_title(data_name)
        ax.set_xlim(0, len(data))
        ax.set_xlabel("Stopping time")
        ax.yaxis.set_ticklabels([])

        plt.tight_layout()

        if bbox_to_anchor is not None:
            ax.legend(
                loc="lower center",
                bbox_to_anchor=bbox_to_anchor,
                ncol=ncol,
            )
        else:
            ax.legend(loc="best")
        # plt.legend(loc="best")
        plt.savefig(
            filename_prefix
            + data_name.replace(" ", "_")
            .replace("(", "")
            .replace(")", "")
            .replace("^\star", "")
            .replace("$", "")
            + ".pdf",
            bbox_inches="tight",
        )
        end_time = time.time()
        info("Finished " + data_name + " in " + str(end_time - start_time) + " seconds")



from setuptools import setup

setup(
    name="rilacs",
    version="1.0.0",
    description="Risk limiting audits via confidence sequences",
    url="http://github.com/WannabeSmith/RiLACS",
    author="Ian Waudby-Smith",
    author_email="ianws@cmu.edu",
    license="BSD 3-Clause",
    packages=["rilacs"],
    install_requires=[
        "confseq",
        "numpy",
        "scipy",
    ],
    zip_safe=False,
)



import math
import numpy as np
from rilacs.strategies import (
    apriori_Kelly_bet,
    get_conv_weights_from_dist,
    square_gamma_dist,
)
from abc import ABC, abstractmethod


class Bettor(ABC):
    def __init__(self):
        self.t = 0

    @abstractmethod
    def update_bet(self, ballot, m_t):
        pass


class AbstractKelly_Bettor(Bettor):
    def __init__(self, conv_weights=np.array([1])):
        super().__init__()
        self.conv_weights = conv_weights

    @abstractmethod
    def update_bet(self, ballot, m_t):
        pass


class DistKelly_Bettor(AbstractKelly_Bettor):
    def __init__(self, dist=square_gamma_dist, D=10, c=1):
        conv_weights = get_conv_weights_from_dist(dist, D=D)
        assert np.abs(np.sum(conv_weights) - 1) < 10e-12
        super().__init__(conv_weights=conv_weights)
        self.c = c

    def update_bet(self, ballot, m_t):
        self.t += 1
        D = len(self.conv_weights)
        d = np.arange(1, D + 1)
        # Create a matrix of bets by taking the
        # outer product between d and 1/((D+1)*m_t)
        with np.errstate(divide="ignore"):
            bets = d[:, None] / ((D + 1) * m_t)
            trunc_bets = np.maximum(0, np.minimum(bets, self.c / m_t))
        return trunc_bets


class Kelly_Bettor(AbstractKelly_Bettor):
    def __init__(self, n_A, n_B, breaks=500, c=1 / 2):
        super().__init__(conv_weights=np.array([1]))

        self.c = c
        self.bets = self._compute_kelly_bets(n_A, n_B, breaks)

    def _compute_kelly_bets(self, n_A, n_B, breaks):
        apriori_bet = apriori_Kelly_bet(n_A=n_A, n_B=n_B)
        bets = np.array([np.repeat(apriori_bet, breaks + 1)])

        return apriori_bet

    def update_bet(self, ballot, m_t):
        self.t += 1
        with np.errstate(divide="ignore"):
            bet = np.maximum(0, np.minimum(self.bets, self.c / m_t))
        return bet


class Audit(ABC):
    def __init__(self, N, alpha=0.05):
        self.l = 0
        self.l_logical = 0
        self.S_t = 0
        self.N = N
        self.alpha = alpha
        self.t = 0

    def _update_l_logical(self, ballot):
        # accumulating S_t instead of S_t / N for numerical stability
        self.S_t += ballot
        self.l_logical = self.S_t / self.N
        return self.l_logical

    @abstractmethod
    def update_cs(self, ballot: float) -> float:
        pass


class Betting_Audit(Audit):
    def __init__(self, N, bettor=DistKelly_Bettor(), breaks=500, alpha=0.05):
        super().__init__(N=N, alpha=alpha)
        self.breaks = breaks
        self.bettor = bettor
        self.m = np.arange(0, 1 + 1 / self.breaks, step=1 / self.breaks)
        self.m_t = np.arange(0, 1 + 1 / self.breaks, step=1 / self.breaks)
        self.capitals = 1

    def _update_m_t(self, ballot):
        if self.N != self.t:
            self.m_t = ((self.N - self.t + 1) * self.m_t - ballot) / (self.N - self.t)
        else:
            # If N == t, then we're at the final step and don't need to take any action
            pass

        return self.m_t

    def update_cs(self, ballot):
        bets = self.bettor.update_bet(ballot, self.m_t)
        self.t += 1
        assert self.t == self.bettor.t

        # Update the capital for each m_t
        # First compute capital process for each of the convex weights
        with np.errstate(invalid="ignore", over="ignore"):
            self.capitals *= np.where(
                self.m_t < 0, math.inf, 1 + bets * (ballot - self.m_t)
            )

        # Take convex combination of capitals to get capital
        convex_weights = self.bettor.conv_weights
        with np.errstate(invalid="ignore"):
            capital = np.sum(convex_weights[:, None] * self.capitals, axis=0)
        # Find the smallest m_t for which the capital is less than 1/alpha,
        # and take a superset
        l_idx = np.where(capital < 1 / self.alpha)[0][0]
        # l_idx = l_idx - 1 if l_idx != 0 else l_idx
        self.l = self.m[l_idx] - 1 / self.breaks

        # Also update logical lower confidence sequence and intersect l with it
        self._update_l_logical(ballot)
        self.l = max(self.l, self.l_logical)

        self._update_m_t(ballot)

        return self.l


class Hoeffding_Audit(Audit):
    def __init__(self, N, alpha=0.05, mu_init=1 / 2):
        super().__init__(N=N, alpha=alpha)
        self.mu = mu_init
        self.margin_numerator = np.log(1 / self.alpha)
        self.margin_denominator = 0
        self.margin = 0
        self.S_tminus1 = 0
        self.mu_numerator = 0
        self.mu_denominator = 0

    def _update_mu(self, ballot):
        l = self._get_lambda(self.t)
        self.mu_numerator += l * (ballot + 1 / (self.N - self.t + 1) * self.S_tminus1)
        self.mu_denominator += l * (1 + (self.t - 1) / (self.N - self.t + 1))
        self.mu = self.mu_numerator / self.mu_denominator

        return self.mu

    def _update_margin(self, ballot):
        l = self._get_lambda(self.t)
        self.margin_numerator += l ** 2 / 8
        self.margin_denominator += l * (1 + (self.t - 1) / (self.N - self.t + 1))
        self.margin = self.margin_numerator / self.margin_denominator
        return self.margin

    def _get_lambda(self, t):
        return np.minimum(np.sqrt(8 * np.log(1 / self.alpha) / t / np.log(t + 1)), 1)

    def _update_S_tminus1(self, ballot):
        self.S_tminus1 += ballot

    def update_cs(self, ballot):
        self.t += 1
        mu = self._update_mu(ballot)
        margin = self._update_margin(ballot)
        l_logical = self._update_l_logical(ballot)

        self.l = np.maximum(mu - margin, l_logical)

        self._update_S_tminus1(ballot)

        return self.l


class Empirical_Bernstein_Audit(Audit):
    def __init__(self, N, alpha=0.05, mu_init=1 / 2):
        super().__init__(N=N, alpha=alpha)
        self.mu = mu_init
        self.margin_numerator = np.log(1 / self.alpha)
        self.margin_denominator = 0
        self.margin = 0
        self.S_tminus1 = 0
        self.mu_numerator = 0
        self.mu_denominator = 0

    def _update_mu(self, ballot):
        l = self._get_lambda(self.t)
        self.mu_numerator += l * (ballot + 1 / (self.N - self.t + 1) * self.S_tminus1)
        self.mu_denominator += l * (1 + (self.t - 1) / (self.N - self.t + 1))
        self.mu = self.mu_numerator / self.mu_denominator

        return self.mu

    def _update_margin(self, ballot):
        l = self._get_lambda(self.t)
        self.margin_numerator += l ** 2 / 8
        self.margin_denominator += l * (1 + (self.t - 1) / (self.N - self.t + 1))
        self.margin = self.margin_numerator / self.margin_denominator
        return self.margin

    def _get_lambda(self, t):
        return np.minimum(np.sqrt(8 * np.log(1 / self.alpha) / t / np.log(t + 1)), 1)

    def _update_S_tminus1(self, ballot):
        self.S_tminus1 += ballot

    def update_cs(self, ballot):
        self.t += 1
        mu = self._update_mu(ballot)
        margin = self._update_margin(ballot)
        l_logical = self._update_l_logical(ballot)

        self.l = np.maximum(mu - margin, l_logical)

        self._update_S_tminus1(ballot)

        return self.l


def lower_cs_from_audit(x: np.ndarray, audit: Audit) -> np.ndarray:
    l = np.zeros(len(x))
    for t in range(len(x)):
        audit.update_cs(x[t])
        l[t] = audit.l
    return l
